@@ -1,4 +1,4 @@
-/* 
+/*
    Unix SMB/CIFS implementation.
    Samba database functions
    Copyright (C) Andrew Tridgell              1999-2000
@@ -10,12 +10,12 @@
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
@@ -61,6 +61,7 @@ enum commands {
 	CMD_NEXT,
 	CMD_SYSTEM,
 	CMD_CHECK,
+	CMD_REPACK,
 	CMD_QUIT,
 	CMD_HELP
 };
@@ -98,6 +99,7 @@ COMMAND_TABLE cmd_table[] = {
 	{"quit",	CMD_QUIT},
 	{"q",		CMD_QUIT},
 	{"!",		CMD_SYSTEM},
+	{"repack",	CMD_REPACK},
 	{NULL,		CMD_HELP}
 };
 
@@ -111,7 +113,7 @@ static void _start_timer(void)
 static double _end_timer(void)
 {
 	gettimeofday(&tp2,NULL);
-	return((tp2.tv_sec - tp1.tv_sec) + 
+	return((tp2.tv_sec - tp1.tv_sec) +
 	       (tp2.tv_usec - tp1.tv_usec)*1.0e-6);
 }
 
@@ -158,7 +160,7 @@ static void print_data(const char *buf,int len)
 		printf("%02X ",(int)((unsigned char)buf[i]));
 		i++;
 		if (i%8 == 0) printf(" ");
-		if (i%16 == 0) {      
+		if (i%16 == 0) {
 			print_asc(&buf[i-16],8); printf(" ");
 			print_asc(&buf[i-8],8); printf("\n");
 			if (i<len) printf("[%03X] ",i);
@@ -166,18 +168,18 @@ static void print_data(const char *buf,int len)
 	}
 	if (i%16) {
 		int n;
-		
+
 		n = 16 - (i%16);
 		printf(" ");
 		if (n>8) printf(" ");
 		while (n--) printf("   ");
-		
+
 		n = i%16;
 		if (n > 8) n = 8;
 		print_asc(&buf[i-(i%16)],n); printf(" ");
 		n = (i%16) - n;
-		if (n>0) print_asc(&buf[i-n],n); 
-		printf("\n");    
+		if (n>0) print_asc(&buf[i-n],n);
+		printf("\n");
 	}
 }
 
@@ -203,6 +205,7 @@ static void help(void)
 "  list                 : print the database hash table and freelist\n"
 "  free                 : print the database freelist\n"
 "  check                : check the integrity of an opened database\n"
+"  repack               : repack the database\n"
 "  speed                : perform speed tests on the database\n"
 "  ! command            : execute system command\n"
 "  1 | first            : print the first record\n"
@@ -257,7 +260,7 @@ static void insert_tdb(char *keyname, size_t keylen, char* data, size_t datalen)
 	dbuf.dptr = (unsigned char *)data;
 	dbuf.dsize = datalen;
 
-	if (tdb_store(tdb, key, dbuf, TDB_INSERT) == -1) {
+	if (tdb_store(tdb, key, dbuf, TDB_INSERT) != 0) {
 		terror("insert failed");
 	}
 }
@@ -284,7 +287,7 @@ static void store_tdb(char *keyname, size_t keylen, char* data, size_t datalen)
 	printf("Storing key:\n");
 	print_rec(tdb, key, dbuf, NULL);
 
-	if (tdb_store(tdb, key, dbuf, TDB_REPLACE) == -1) {
+	if (tdb_store(tdb, key, dbuf, TDB_REPLACE) != 0) {
 		terror("store failed");
 	}
 }
@@ -306,11 +309,11 @@ static void show_tdb(char *keyname, size_t keylen)
 	    terror("fetch failed");
 	    return;
 	}
-	
+
 	print_rec(tdb, key, dbuf, NULL);
-	
+
 	free( dbuf.dptr );
-	
+
 	return;
 }
 
@@ -354,23 +357,23 @@ static void move_rec(char *keyname, size_t keylen, char* tdbname)
 		terror("fetch failed");
 		return;
 	}
-	
+
 	print_rec(tdb, key, dbuf, NULL);
-	
+
 	dst_tdb = tdb_open(tdbname, 0, 0, O_RDWR, 0600);
 	if ( !dst_tdb ) {
 		terror("unable to open destination tdb");
 		return;
 	}
-	
-	if ( tdb_store( dst_tdb, key, dbuf, TDB_REPLACE ) == -1 ) {
+
+	if (tdb_store( dst_tdb, key, dbuf, TDB_REPLACE ) != 0) {
 		terror("failed to move record");
 	}
 	else
 		printf("record moved\n");
-	
+
 	tdb_close( dst_tdb );
-	
+
 	return;
 }
 
@@ -409,16 +412,19 @@ static int traverse_fn(TDB_CONTEXT *the_tdb, TDB_DATA key, TDB_DATA dbuf, void *
 
 static void info_tdb(void)
 {
-	int count;
-	total_bytes = 0;
-	if ((count = tdb_traverse(tdb, traverse_fn, NULL)) == -1)
+	char *summary = tdb_summary(tdb);
+
+	if (!summary) {
 		printf("Error = %s\n", tdb_errorstr(tdb));
-	else
-		printf("%d records totalling %d bytes\n", count, total_bytes);
+	} else {
+		printf("%s", summary);
+		free(summary);
+	}
 }
 
 static void speed_tdb(const char *tlimit)
 {
+	const char *str = "store test", *str2 = "transaction test";
 	unsigned timelimit = tlimit?atoi(tlimit):0;
 	double t;
 	int ops;
@@ -430,9 +436,9 @@ static void speed_tdb(const char *tlimit)
 	do {
 		long int r = random();
 		TDB_DATA key, dbuf;
-		key.dptr = (unsigned char *)"store test";
+		key.dptr = discard_const_p(uint8_t, str);
 		key.dsize = strlen((char *)key.dptr);
-		dbuf.dptr = (unsigned char *)&r;
+		dbuf.dptr = (uint8_t *) &r;
 		dbuf.dsize = sizeof(r);
 		tdb_store(tdb, key, dbuf, TDB_REPLACE);
 		t = _end_timer();
@@ -444,12 +450,9 @@ static void speed_tdb(const char *tlimit)
 	printf("Testing fetch speed for %u seconds\n", timelimit);
 	_start_timer();
 	do {
-		long int r = random();
-		TDB_DATA key, dbuf;
-		key.dptr = (unsigned char *)"store test";
+		TDB_DATA key;
+		key.dptr = discard_const_p(uint8_t, str);
 		key.dsize = strlen((char *)key.dptr);
-		dbuf.dptr = (unsigned char *)&r;
-		dbuf.dsize = sizeof(r);
 		tdb_fetch(tdb, key);
 		t = _end_timer();
 		ops++;
@@ -462,9 +465,9 @@ static void speed_tdb(const char *tlimit)
 	do {
 		long int r = random();
 		TDB_DATA key, dbuf;
-		key.dptr = (unsigned char *)"transaction test";
+		key.dptr = discard_const_p(uint8_t, str2);
 		key.dsize = strlen((char *)key.dptr);
-		dbuf.dptr = (unsigned char *)&r;
+		dbuf.dptr = (uint8_t *) &r;
 		dbuf.dsize = sizeof(r);
 		tdb_transaction_start(tdb);
 		tdb_store(tdb, key, dbuf, TDB_REPLACE);
@@ -517,7 +520,7 @@ static void first_record(TDB_CONTEXT *the_tdb, TDB_DATA *pkey)
 {
 	TDB_DATA dbuf;
 	*pkey = tdb_firstkey(the_tdb);
-	
+
 	dbuf = tdb_fetch(the_tdb, *pkey);
 	if (!dbuf.dptr) terror("fetch failed");
 	else {
@@ -529,17 +532,17 @@ static void next_record(TDB_CONTEXT *the_tdb, TDB_DATA *pkey)
 {
 	TDB_DATA dbuf;
 	*pkey = tdb_nextkey(the_tdb, *pkey);
-	
+
 	dbuf = tdb_fetch(the_tdb, *pkey);
-	if (!dbuf.dptr) 
+	if (!dbuf.dptr)
 		terror("fetch failed");
 	else
 		print_rec(the_tdb, *pkey, dbuf, NULL);
 }
 
-static int count(TDB_DATA key, TDB_DATA data, void *private)
+static int count(TDB_DATA key, TDB_DATA data, void *private_data)
 {
-	(*(unsigned int *)private)++;
+	(*(unsigned int *)private_data)++;
 	return 0;
 }
 
@@ -608,6 +611,10 @@ static int do_command(void)
 			bIterate = 0;
 			tdb_transaction_commit(tdb);
 			return 0;
+		case CMD_REPACK:
+			bIterate = 0;
+			tdb_repack(tdb);
+			return 0;
 		case CMD_TRANSACTION_CANCEL:
 			bIterate = 0;
 			tdb_transaction_cancel(tdb);
@@ -690,7 +697,7 @@ static int do_command(void)
 	return 0;
 }
 
-static char *convert_string(char *instring, size_t *sizep)
+static char *tdb_convert_string(char *instring, size_t *sizep)
 {
 	size_t length = 0;
 	char *outp, *inp;
@@ -756,15 +763,15 @@ int main(int argc, char *argv[])
 					}
 				}
 			}
-			if (arg1) arg1 = convert_string(arg1,&arg1len);
-			if (arg2) arg2 = convert_string(arg2,&arg2len);
+			if (arg1) arg1 = tdb_convert_string(arg1,&arg1len);
+			if (arg2) arg2 = tdb_convert_string(arg2,&arg2len);
 			if (do_command()) break;
 		}
 		break;
 	case 5:
-		arg2 = convert_string(argv[4],&arg2len);
+		arg2 = tdb_convert_string(argv[4],&arg2len);
 	case 4:
-		arg1 = convert_string(argv[3],&arg1len);
+		arg1 = tdb_convert_string(argv[3],&arg1len);
 	case 3:
 		cmdname = argv[2];
 	default:
